import sys
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from PyQt5.QtWidgets import QApplication, QMainWindow, QVBoxLayout, QWidget

import WinDeklar.signal_aux as sg


class RealTimeGraph:
    def __init__(self, fig, data_provider, min_x=-1, repeat_length=50, frames=None, interval=100):
        self.repeat_length = repeat_length
        self.data_provider = data_provider if isinstance(data_provider, list) else [data_provider]

        # Create an axis in the given figure
        self.ax = fig.add_subplot(111)
        self.lines = []
        for dp in self.data_provider:
            line, = self.ax.plot([], [], color=dp.color)
            self.lines.append([line, dp, sg.SignalHistory(self.repeat_length), sg.SignalHistory(self.repeat_length)])

        # Set the axis limits
        self.ax.set_xlim(min_x, self.repeat_length)
        min_y = None
        max_y = None
        for dp in self.data_provider:
            min_y1, max_y1 = dp.get_bounds()
            if min_y is None or min_y1 < min_y:
                min_y = min_y1
            if max_y is None or max_y1 > max_y:
                max_y = max_y1
        self.ax.set_ylim(min_y, max_y)

        self.anim = FuncAnimation(fig, self.update_frame, frames=frames, interval=interval, blit=False)

    def update_frame(self, frame):
        for [line, dp, xs, ys] in self.lines:
            x, y = dp.get_next_values(frame)
            # print(' frame:%s x:%s y:%s' % (frame, x, y))
            xs.append(x)
            ys.append(y)

            x_max = xs.max()
            if x_max > self.repeat_length:
                self.ax.set_xlim(x_max - self.repeat_length, x_max)
            line.set_data(xs.values, ys.values)


class RealTimeDataProvider(object):
    def __init__(self, min_y=0.0, max_y=10.0, color='Red'):
        self.min_y = min_y
        self.max_y = max_y
        self.color = color

    def get_bounds(self):
        return self.min_y, self.max_y

    def get_min_y(self):
        return self.min_y

    def get_max_y(self):
        return self.max_y

    def get_next_values(self, i):
        return i, i


class RealTimeRandomDataProvider(RealTimeDataProvider):
    def __init__(self, min_y=0.0, max_y=10.0, color='Red'):
        super(RealTimeRandomDataProvider, self).__init__(min_y=min_y, max_y=max_y, color=color)

    def get_next_values(self, i):
        return i, np.random.uniform(self.min_y, self.max_y)


class RealTimeFunctionDataProvider(RealTimeDataProvider):
    def __init__(self, min_y=-1.2, max_y=1.2,function=np.sin, inc=np.radians(10), color='Red'):
        self.function = function
        self.inc      = inc
        self.last_r   = 0.0
        super(RealTimeFunctionDataProvider, self).__init__(min_y=min_y, max_y=max_y, color=color)

    def get_next_values(self, i):
        y = self.function(self.last_r)
        self.last_r += self.inc
        return i, y


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__(None)
        self.setWindowTitle("Animated Graph")

        # Create a Matplotlib figure
        fig = plt.figure()

        # Create an instance of CustomAnimation and pass the figure
        provider = [RealTimeFunctionDataProvider(function=np.sin, color='Blue'),
                    RealTimeFunctionDataProvider(function=np.cos, color='Red'),
                    RealTimeRandomDataProvider(min_y=-1, max_y=1, color='Orange')]
        self.animation = RealTimeGraph(fig, provider)

        # Create a Matplotlib canvas to display the animation
        canvas = FigureCanvas(fig)

        # Add the canvas to the vertical layout
        layout = QVBoxLayout()
        layout.addWidget(canvas)

        # Create a widget to hold the layout
        widget = QWidget(None)
        widget.setLayout(layout)

        # Set the widget as the central content of the main window
        self.setCentralWidget(widget)

        # Start the animation
        # self.animation._start()

        # Show the main window
        self.show()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = MainWindow()
    sys.exit(app.exec_())



class PointsBox:
    """
    Keeps an AABB that contains a set of points
       the box is creating adding points, but the points itself are not stored, just the min and max values
       of each coordinate
       AABB: Axis Aligned Bounding Box
    """

    def __init__(self, min_x=0.0, max_x=0.0, min_y=0.0, max_y=0.0):
        self.is_empty = True
        self.min_x = min_x
        self.max_x = max_x
        self.min_y = min_y
        self.max_y = max_y

    def reset(self):
        self.is_empty = True

    def size(self):
        return self.min_x, self.max_x, self.min_y, self.max_y

    def add_points(self, points):
        """
        Use the points to update the box size
        :param points: :type list of [x, y]
        :return: nothing
        """
        for p in points:
            self.add_point(p)

    def add_point(self, p):
        [x, y] = p
        if self.is_empty:
            self.min_x, self.max_x, self.min_y, self.max_y = [x, x, y, y]
            self.is_empty = False
        else:
            self.min_x, self.max_x = update_bounds(x, self.min_x, self.max_x)
            self.min_y, self.max_y = update_bounds(y, self.min_y, self.max_y)

    def set_lim(self, ax, inc=1.1):
        min1 = min(self.min_x, self.min_y) * inc
        max1 = max(self.max_x, self.max_y) * inc
        ax.set_xlim([min1, max1])
        ax.set_ylim([min1, max1])

    def set_bounds(self, ax, inc=1.0):
        if self.is_empty:
            # defensive programming
            return
        ax.set_xbound(lower=add_inc(self.min_x, inc, -1), upper=add_inc(self.max_x, inc, 1))
        ax.set_ybound(lower=add_inc(self.min_y, inc, -1), upper=add_inc(self.max_y, inc, 1))

    def __str__(self):
        return 'min_x:%.2f max_x:%.2f min_y:%.2f max_y:%.2f' % (self.min_x, self.max_x, self.min_y, self.max_y)


def update_bounds(x, min_x, max_x):
    new_min_x = x if x < min_x else min_x
    new_max_x = x if x > max_x else max_x
    return new_min_x, new_max_x


def add_inc(value, inc, sign):
    """
    Returns a value bigger (by inc percentage) than itself taking in count the sign (negative must be less, positive
    greater
    :param value:
    :param inc:
    :param sign:
    :return:
    """
    delta = abs(value)*(inc - 1)
    return value + sign*delta



#!/usr/bin/env python

import os
import yaml
import json


def get_yaml_file(file_name, directory='', type='r', must_exist=True, verbose=False):
    if not file_name:
        print('No file name given')
        return
    if os.path.isabs(file_name):
        full_file_name = file_name
    else:
        if directory is None:
            script_dir = ''
        elif directory == '':
            script_dir = os.path.dirname(__file__) + '/'  # <-- absolute dir the script is in
        else:
            script_dir = directory + '/'
        full_file_name = script_dir + file_name

    try:
        with open(full_file_name, type) as yml_file:
            if verbose:
                print('Loading %s ...' % full_file_name)
            cfg = yaml.load(yml_file, Loader=yaml.FullLoader)
            if verbose:
                print('loaded')
    except IOError:
        cfg = {}
        if must_exist:
            raise Exception('File %s not found' % full_file_name)
    return cfg


def get_json_file(file_name, directory='', type='r', must_exist=True, verbose=False):
    if not file_name:
        print('No file name given')
        return
    if os.path.isabs(file_name):
        full_file_name = file_name
    else:
        if directory is None:
            script_dir = ''
        elif directory == '':
            script_dir = os.path.dirname(__file__) + '/'  # <-- absolute dir the script is in
        else:
            script_dir = directory + '/'
        full_file_name = script_dir + file_name

    try:
        with open(full_file_name, type) as json_file:
            if verbose:
                print('Loading %s ...' % full_file_name)
            cfg = json.load(json_file)  # Loader=yaml.FullLoader
            if verbose:
                print('loaded')
    except IOError:
        cfg = {}
        if must_exist:
            raise Exception('File %s not found' % full_file_name)
    return cfg


def save_yaml_file(dictionary, file_name, directory='', verbose=False):
    if not file_name:
        print('No file name given')
        return
    elif directory is None:
        script_dir = ''
    elif directory == '':
        script_dir = os.path.dirname(__file__)  # <-- absolute dir the script is in
    else:
        script_dir = directory
    full_file_name = script_dir + '/' + file_name

    with open(full_file_name, 'w') as f:
        yaml.dump(dictionary, f, default_flow_style=False)
        if verbose:
            print('saved to: %s' % full_file_name)
    return full_file_name


def save_json_file(dictionary, file_name, directory='', verbose=False):
    if not file_name:
        print('No file name given')
        return
    elif directory is None:
        script_dir = ''
    elif directory == '':
        script_dir = os.path.dirname(__file__)  # <-- absolute dir the script is in
    else:
        script_dir = directory
    full_file_name = script_dir + '/' + file_name

    with open(full_file_name, 'w') as f:
        json.dump(dictionary, f)
        if verbose:
            print('saved to: %s' % full_file_name)
    return full_file_name


def get_record(all_file, name, collection_name, record_name, key_name='name', alternative_key_name='desc'):
    # give a yaml file with a collection in it, returns one element of the collection with a given name
    # ex:
    #      tests:             # collection name
    #         - test:         # record name
    #             name:  xxx  # key_name: name
    if collection_name not in all_file:
        raise Exception('Collection %s not found in %s' % (collection_name, all_file))

    group_data = all_file[collection_name]
    for test_data in group_data:
        data = test_data[record_name]
        # print(data)
        if key_name in data:
            valid_key_name = key_name
        elif alternative_key_name in data:
            valid_key_name = alternative_key_name
        else:
            raise Exception('Nor key %s or %s present' % (key_name, alternative_key_name))
        if data[valid_key_name] == name:
            return data
    raise Exception('Name %s not found in %s' % (name, record_name))


def get_group_data(file_name, group_name, name='', key_name='name'):
    # Given a yaml FileName a group_name and a name, returns its data
    # ex: test.yaml, 'test', 'my test' return a group name 'test' in file test.yaml whose name is 'my test'
    #   if name is '' returns the first one
    #   'name' is configurable with key_name (ex: searching for 'external_name:' instead of 'name:'

    all_file = get_yaml_file(file_name)
    for definition in all_file:
        if group_name in definition:
            kvs = definition[group_name]
            if name == '':
                return kvs
            for key, value in kvs.items():
                if key == key_name and value == name:
                    return kvs

    not_found = '"- %s" group' % group_name if name == '' else '"- %s" named "%s"' % (group_name, name)
    raise Exception('%s does not exists in %s' % (not_found, file_name))


def get_all_names(file_name, group_name, key_name='name', key_description='description'):
    # given a file_name and group_name returns all the groups' name
    names = []
    all_file = get_yaml_file(file_name)
    for definition in all_file:
        if group_name in definition:
            kvs = definition[group_name]
            if key_name in kvs:
                desc = kvs[key_description] if key_description in kvs else ''
                names.append((kvs[key_name], desc))
    return names


def get_file_for_write(file_name, type='w+', directory=None):
    full_name = '%s/%s' % (directory, file_name) if directory is not None else file_name
    return open(full_name, type)


def string_to_dict(string):
    return yaml.safe_load(string)


def file_name_extension(file_name):
    """
    Returns the extension of the file name
    :param file_name:
    :return:
    """
    return os.path.splitext(file_name)[-1].lower()


def file_name_without_extension(file_name):
    """
    Returns the file name without the extension
    :param file_name:
    :return:
    """
    return os.path.splitext(file_name)[0] if file_name is not None else None


def directory_path(file_name):
    """
    Returns the directory path of a full file name
    :param file_name:
    :return:
    """
    return os.path.dirname(file_name)


def get_file_name_with_other_extension(file_path, ext='yaml'):
    dir_name      = os.path.dirname(os.path.abspath(file_path))
    base_name     = file_name_without_extension(os.path.basename(file_path))
    new_file_name = '%s/%s.%s' % (dir_name, base_name, ext)
    return new_file_name



#!/usr/bin/env python

import sys
import numpy as np

import WinDeklar.QTAux as QTAux
import WinDeklar.WindowForm as WinForm
import WinDeklar.graph_aux as ga


class ExampleHost(WinForm.HostModel):
    """
    Shows tentacles
    """

    def __init__(self, function=np.sin, color='Blue', y_bounds=(-1.2, 1.2)):
        # keys (names used in the yaml definition file)
        self.start_stop_key        = 'start_stop'
        self.start_stop_action_key = 'start_stop_action'

        # particular data
        self.function = function
        self.color    = color
        self.y_bounds = y_bounds

        initial_values = {}
        super(ExampleHost, self).__init__(initial_values=initial_values)

    def get_data_provider(self, interval=100, min_x=0.0, max_x=10.0, data_provider=None):
        """
        Returns the basic setup of the graph, most important one is the data provider
        :param interval:       time in milliseconds to show each point in the graph
        :param min_x:          start value in the x-axis
        :param max_x:          max value in the x-axis, after that it starts to scroll left
        :param data_provider:  subclass of RealTimeDataProvider
        :return:
        """
        dt             = interval/1000  # seconds
        max_points     = int(max_x/dt) + 1
        x_bounds       = [min_x, max_x]
        data_provider1 = [ga.RealTimeFunctionDataProvider(dt=dt, min_y=self.y_bounds[0], max_y=self.y_bounds[1],
                                                          function=self.function, color=self.color)]
        return interval, max_points, x_bounds, data_provider1

    def start_stop_animation(self):
        """
        Logit to start/stop the graph with only one button (who changes its title depending on the graph is
        running or not_
        :return:
        """
        if self.anim_is_running():
            self.stop_animation()
            self.set_control_title(self.start_stop_key, 'Restart Animation')
            self.set_control_title(self.start_stop_action_key, 'Restart Animation')
        else:
            self.start_animation()
            self.set_control_title(self.start_stop_key, 'Pause Animation')
            self.set_control_title(self.start_stop_action_key, 'Pause Animation')


if __name__ == '__main__':
    app = QTAux.def_app()
    provider = ExampleHost()        # class to handle events
    WinForm.set_winform(__file__, provider)
    sys.exit(app.exec_())



from WinDeklar.WindowForm import SimpleFigure
import WinDeklar.graph_aux as ga


def show_simple_window(title='Sine and Cosine graph', number_of_points=100, size=(5, 2),
                       functions=(('Sine', 'Blue'), ('Cosine', 'Red'))):
    window = SimpleFigure(title=title, size=size, adjust_size=False)
    for [function_name, color] in functions:
        graph_one_function(function_name, window.ax, number_of_points, color)
    window.show()


def graph_one_function(function_name, ax, number_of_points, color):
    points, msg = ga.graph_points_for_many_functions(function_name, number_of_points)
    if points is None:
        print('%s not implemented' % function_name)
        return
    ga.graph_points(ax, points, color=color)


if __name__ == '__main__':
    show_simple_window()



#!/usr/bin/env python

from __future__ import print_function

from datetime import datetime

import WinDeklar.yaml_functions as yaml


class Record:
    def __init__(self, file_name, dir='/tmp', indent=2, ext='yaml', add_time_stamp=True):
        self.file          = None       # only use file if something is going to be writing
        self.file_name     = file_name
        self.ext           = ext
        self.dir           = dir
        self.indent        = indent     # number of characters to indent in each level
        self.indent_spaces = ''
        self.add_ts        = add_time_stamp
        for _ in range(0, self.indent):
            self.indent_spaces += ' '

        self.file_name_ts = self.get_file_name_with_time_stamp()

        self.groups = {}
        self.opened = False

    def write_group(self, group_name, values, level=0, is_array=False):
        if values is None:
            # do nothing
            return
        if is_array:
            self.write_group_header(group_name, level)
        array_str = '- ' if is_array else ''
        front_str = level_spaces(level, self.indent_spaces)
        self.write_ln(front_str + array_str + group_name + ':')
        sub_spaces = front_str + self.indent_spaces
        if is_array:
            sub_spaces += self.indent_spaces
        for k, v in values.items():
            self.write_ln('%s%s: %s' % (sub_spaces, k, v))

    def write_group_header(self, group_name, level):
        if group_name not in self.groups:
            front_str = level_spaces(level - 1, self.indent_spaces)
            self.write_ln(front_str + group_name + 's:')
            self.groups[group_name] = True

    def write_ln(self, string):
        if not self.opened:
            self.file   = yaml.get_file_for_write(self.get_full_file_name())
            self.opened = True
            self.groups = {}
        self.file.write('%s\r\n' % string)

    def get_file_name_with_time_stamp(self):
        directory = '%s/' % self.dir if self.dir is not None else ''
        full_file_name = "%s%s" % (directory, self.file_name)
        if self.add_ts:
            now        = datetime.now()
            now_string = now.strftime("%Y_%m_%d_%H_%M_%S")  # ""%m/%d/%Y, %H:%M:%S"))
            full_file_name += '_%s' % now_string
        return full_file_name

    def get_full_file_name(self, with_ext=True):
        if self.file_name_ts is None:
            # get a new file name (needed when recording a behavior with more than 1000 steps because it is writen in
            # chunks)
            self.file_name_ts = self.get_file_name_with_time_stamp()
        if with_ext and yaml.file_name_extension(self.file_name) == '':
            # check if ext is present, if not add it
            full_file_name = '%s.%s' % (self.file_name_ts, self.ext)
        else:
            full_file_name = self.file_name_ts
        return full_file_name

    def close(self):
        if self.opened:
            self.opened = False
            self.file.close()
            self.file_name_ts = None


def level_spaces(level, indent_spaces):
    spaces = ''
    for _ in range(1, level+1):
        spaces += indent_spaces
    return spaces


if __name__ == "__main__":

    # just for testing
    r = Record('record_test.yaml')
    r.write_group('vehicle', {'l': 0.93, 'w': 0.7})
    r.write_ln('cycles:')
    r.write_group('cycle', {'age': 1, 'v': 2}, level=1, is_array=True)
    r.write_group('cycle', {'age': 2, 'v': 3}, level=1, is_array=True)



#!/usr/bin/env python
import functools
import sys
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib import animation
from PyQt5 import QtGui, QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

import WinDeklar.points_box as pb
import WinDeklar.QTAux as QTAux
import WinDeklar.record as rc
import WinDeklar.signal_aux as sg
import WinDeklar.yaml_functions as yaml


def set_winform(file_path, provider, ext='yaml'):
    win_config_name = yaml.get_file_name_with_other_extension(file_path, ext)
    return ConfigurableWindow(win_config_name, provider)


class ConfigurableWindow(QtWidgets.QMainWindow):
    """
    Defines a Window with many components inside in a declarative manner in a config file
    Notes:
        - a config_file_name is a yaml that has the definition (see view_map.yaml)
        - a typical windows has some controls in the left and a Figure in the right (like an ego view or a map)
    """

    def __init__(self, config_file_name, provider):
        """
        Init
        :param config_file_name: name of the file containing the window definition
        :param provider: handles all the form logic, usually a subclass of HostModel
        """
        super(ConfigurableWindow, self).__init__(parent=None)

        self.controls = []
        self.provider = provider
        self.provider.set_main_window(self)
        self.fig_view = None

        self.win_config = get_win_config(config_file_name)

        # status bar logic
        self.statusbar    = create_status_bar(self.win_config, self)
        self.progress_bar = GeneralProgressBar(widget=self.statusbar, stretch=1, visible=False)

        # menu bar
        menu_bar = create_menu_bar(self.win_config, self, self.provider)
        if menu_bar is not None:
            self.setMenuBar(menu_bar)

        # toolbar
        toolbar = create_toolbar(self.win_config, self, self.provider, self.controls)
        if toolbar is not None:
            self.addToolBar(toolbar)

        # Define the geometry of the main window
        size = self.win_config.get('size', [300, 300, 300, 300])
        QTAux.set_window(self, get_title(self.win_config, self.provider), size)

        # Create FRAME
        self.FRAME = QtWidgets.QFrame(self)
        if 'back_color' in self.win_config:
            [c1, c2, c3, c4] = self.win_config['back_color']
            self.FRAME.setStyleSheet("QWidget { background-color: %s }" % QtGui.QColor(c1, c2, c3, c4).name())
        self.LAYOUT   = QtWidgets.QGridLayout()
        self.fig_view = set_layout(self.LAYOUT, self.win_config.get('layout', []), self.controls, self, row_col=[0, 0])
        self.FRAME.setLayout(self.LAYOUT)
        self.setCentralWidget(self.FRAME)

        self.provider.initialize()
        if self.fig_view:
            self.fig_view.update_figure()
        self.show()

    def get_control_value(self, name):
        # print('get value for %s' % name)
        return self.provider.get_control_value(name)

    def control_has_value(self, name):
        return self.provider.control_has_value(name)

    def set_control_value(self, name, value):
        self.provider.set_control_value(name, value)
        self.refresh_other_widgets(name)
        self.refresh()

    def set_control_min_max(self, control_name, min_value, max_value):
        control = self.get_control_by_name(control_name)
        if control is None:
            return
        control.set_min_max(min_value, max_value)

    def set_control_title(self, control_name, new_title):
        control = self.get_control_by_name(control_name)
        if control is None:
            return
        control.set_ename(new_title)

    def refresh(self):
        if self.fig_view:
            # initial values can be set before fig_view was created
            self.fig_view.update_figure()

    def refresh_controls(self):
        for control in self.controls:
            control.refresh()

    def refresh_control(self, control_name):
        control = self.get_control_by_name(control_name)
        if control is not None:
            control.refresh()

    def refresh_other_widgets(self, control_name):
        # if a widget that affect the value of others changed, it is needed to refresh all others'
        #    ex: in view_color, the encoding combo changes the values of low and high sliders
        changed = self.get_control_by_name(control_name)
        if changed is not None and changed.refresh_others:
            # print('refresh others')
            for control in self.controls:
                if control != changed:
                    control.refresh()

    def get_control_by_name(self, control_name):
        for control in self.controls:
            if control.name == control_name:
                return control
        return None

    def show_status_bar_msg(self, msg):
        if self.statusbar is None:
            print('No status bar defined')
            return
        self.statusbar.showMessage(msg)


class FigureView(FigureCanvas):
    """
    Display a drawing that responds to a change in control values
    """

    def __init__(self, parent, size=(1, 1), axes_limits=None, subtype=None, scaled=True, x_visible=True,
                 y_visible=True, animation_key='animation'):

        self.parent   = parent
        self.size_dim = size
        self.subtype  = subtype
        self.anim     = None
        width, height = self.size_dim
        self.figure   = Figure(figsize=None)  # not necessary to set figsize
        self.axes     = self.figure.add_subplot(111)

        if axes_limits is None:
            self.x_lower, self.x_upper = [-width, width]
            self.y_lower, self.y_upper = [-height, height]
        else:
            [self.x_lower, self.x_upper, self.y_lower, self.y_upper] = axes_limits

        self.x_visible, self.y_visible = [x_visible, y_visible]
        self.scaled = False if self.subtype == animation_key else scaled

        self.set_axis()

        FigureCanvas.__init__(self, self.figure)
        self.setParent(self.parent)

        FigureCanvas.setSizePolicy(self, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        self.figure.tight_layout()

        self.figure.canvas.mpl_connect('button_press_event', self.onclick)
        self.figure.canvas.mpl_connect('motion_notify_event', self.on_mouse_move)

        dec         = 0.95
        self.text_x = - width * dec
        self.text_y = height  * dec

        # animation logic
        self.graph_lines   = None  # to signal graph are not initialized yet
        self.data_provider = None
        self.graph_bounds  = None
        self.points_in_graph = 0
        self.anim_is_running = False
        if self.subtype == animation_key:
            interval, self.points_in_graph, self.graph_bounds, self.data_provider = \
                self.parent.provider.get_data_provider()
            if self.data_provider is None:
                raise Exception(
                    'Figure is defined as "animation" but not data provider is given, implement get_data_provider() '
                    'in provider ')
            self.anim = animation.FuncAnimation(self.figure, self.update_frame, frames=None, interval=interval,
                                                blit=False)

    def clear(self):
        if self.anim is not None:
            # in case of animation do not change axes limits, anim itself does it
            return

        self.axes.clear()
        self.set_axis()

    def set_axis(self):
        if self.anim is not None:
            # in case of animation do not change axes limits, anim itself does it
            return

        if self.scaled:
            self.axes.axis('scaled')

        axes_limits = self.parent.provider.axes_limits()
        if axes_limits is not None:
            self.x_lower, self.x_upper, self.y_lower, self.y_upper = axes_limits
        self.axes.set_xbound(lower=self.x_lower, upper=self.x_upper)
        self.axes.set_ybound(lower=self.y_lower, upper=self.y_upper)
        self.axes.get_xaxis().set_visible(self.x_visible)
        self.axes.get_yaxis().set_visible(self.y_visible)
        self.axes.set_ylabel(' ', fontsize=20)  # quick fix to assure y values fit in the figure

    def onclick(self, event):
        self.parent.provider.on_mouse_click(event, self.axes, self)
        self.set_axis()
        self.draw()

    def on_mouse_move(self, event):
        if self.parent.provider.on_mouse_move(event, self.axes):
            self.set_axis()
            self.draw()

    def popup_context_menu(self, actions=(), update_figure=True):
        """
        Displays a contex menu
        :param actions: list of [name, event]
        :param update_figure: whether update the underlying figure after showing the menu, useful when an action
                              changes something, ex: in view_build_map changing the Wall position
        :return:
        """
        if len(actions) == 0:
            return
        context_menu = QTAux.Menu(self, actions=actions)
        context_menu.popup()
        if update_figure:
            self.update_figure()

    def update_figure(self):
        self.clear()
        self.parent.provider.update_view(self.axes)
        self.parent.provider.apply_zoom()
        self.draw()

    def get_control_value(self, name):
        return self.parent.provider.get_control_value(name)

    def control_has_value(self, name):
        return self.parent.provider.control_has_value(name)

    def set_control_value(self, name, value):
        # print('%s changed to %s' % (name, value))
        self.parent.provider.set_control_value(name, value)
        self.parent.refresh_other_widgets(name)
        self.update_figure()

    # animation methods
    def update_frame(self, frame_number):
        # first time do initializations
        if self.graph_lines is None:
            self.initialize_graph_lines(self.graph_bounds, self.data_provider)

        self.anim_is_running = True
        # update graphs
        for [line, dp, xs, ys] in self.graph_lines:
            x, y = dp.get_next_values(frame_number)
            # print(' frame:%s x:%s y:%s' % (frame, x, y))
            xs.append(x)
            ys.append(y)

            x_max = xs.max()
            if x_max > self.graph_bounds[1]:
                self.axes.set_xlim(x_max - self.graph_bounds[1], x_max)
            line.set_data(xs.values, ys.values)

    def initialize_graph_lines(self, bounds, data_provider):
        """
        Initialize each of the graph lines
        :param bounds:
        :param data_provider:
        :return:
        """
        # Create a line for each data provider
        self.graph_lines = []
        for dp in data_provider:
            line, = self.axes.plot([], [], color=dp.color)
            xs = sg.SignalHistory(self.points_in_graph)
            ys = sg.SignalHistory(self.points_in_graph)
            self.graph_lines.append([line, dp, xs, ys])

        # Set the axis limits
        min_x, max_x     = bounds
        self.axes.set_xlim(min_x, max_x)
        min_y = None
        max_y = None
        for dp in self.data_provider:
            min_y1, max_y1 = dp.get_bounds()
            if min_y is None or min_y1 < min_y:
                min_y = min_y1
            if max_y is None or max_y1 > max_y:
                max_y = max_y1
        self.axes.set_ylim(min_y, max_y)

    def stop_animation(self):
        if self.anim is None:
            return
        self.anim.event_source.stop()
        self.anim_is_running = False

    def start_animation(self):
        if self.anim is None:
            return
        self.anim.event_source.start()
        self.anim_is_running = True


class SimpleFigure:
    """
    Provides the functionality to display a Window with a Figure inside
    Used mainly for displaying tests
    """

    def __init__(self, title='Figure', scaled=True, inc=1.0, size=(1, 1), adjust_size=True):
        plt.figure(title, figsize=size)
        self.ax = plt.subplot(111)
        if scaled:
            self.ax.axis('scaled')
        self.ax.set_xlim([-size[0], size[0]])
        self.ax.set_ylim([-size[1], size[1]])
        self.inc        = inc
        self.set_bounds = adjust_size
        self.box_size   = pb.PointsBox()

    def resize(self, points):
        self.box_size.add_points(points)

    def show(self):
        if self.set_bounds:
            self.box_size.set_bounds(self.ax, inc=self.inc)
        plt.show()


class Dialog(QtWidgets.QDialog):
    """
    Functionality for an Input Panel
    """
    def __init__(self, dialog_name, provider, default_size=(300, 300, 400, 400)):
        super(Dialog, self).__init__(parent=None)

        self.controls = []
        self.provider = provider
        self.provider.set_main_window(self)
        self.fig_view = None

        self.win_config = get_win_config(dialog_name)

        # Define the geometry of the main window
        size = self.win_config.get('size', default_size)
        QTAux.set_window(self, get_title(self.win_config, self.provider), size)

        # Create FRAME
        self.FRAME = QtWidgets.QFrame(self)
        self.LAYOUT   = QtWidgets.QGridLayout()
        self.fig_view = set_layout(self.LAYOUT, self.win_config.get('layout', []), self.controls, self, row_col=[0, 0])
        self.FRAME.setLayout(self.LAYOUT)

        self.provider.initialize()
        if self.fig_view:
            self.fig_view.update_figure()

    def show(self):
        self.exec_()

    def get_control_value(self, name):
        # print('get value for %s' % name)
        return self.provider.get_control_value(name)

    def control_has_value(self, name):
        return self.provider.control_has_value(name)

    def set_control_value(self, name, value):
        # print('%s changed to %s' % (name, value))
        self.provider.set_control_value(name, value)
        self.refresh_other_widgets(name)
        self.refresh()

    def refresh(self):
        if self.fig_view:
            # initial values can be set before fig_view was created
            self.fig_view.update_figure()

    def refresh_controls(self):
        for control in self.controls:
            control.refresh()

    def refresh_control(self, control_name):
        control = self.get_control_by_name(control_name)
        if control is not None:
            control.refresh()

    def refresh_other_widgets(self, control_name):
        # if a widget that affect the value of others changed, it is needed to refresh all others'
        #    ex: in view_color, the encoding combo changes the values of low and high sliders
        changed = self.get_control_by_name(control_name)
        if changed is not None and changed.refresh_others:
            # print('refresh others')
            for control in self.controls:
                if control != changed:
                    control.refresh()

    def get_control_by_name(self, control_name):
        for control in self.controls:
            if control.name == control_name:
                return control
        return None

    def confirmed(self):
        print('confirmed')
        self.close()


class HostModel(object):
    """
    Host for ConfigurableWindow (and ControlAndFigureWindow)
       provides all the methods need it by ConfigurableWindow (like state management and control's definition)
    """

    def __init__(self, initial_values=None):
        # keys
        self.zoom_key = 'zoom'

        self.state       = initial_values if initial_values is not None else {}
        self.main_window = None
        self.box_size    = pb.PointsBox()

        self.zoom_center = None   # point where to center Zoom
        self.zoom_radius = 10.0   # radius around

    def set_main_window(self, main_window):
        self.main_window = main_window

    def refresh(self):
        if self.main_window is None:
            return
        self.main_window.refresh_controls()
        self.main_window.refresh()

    def refresh_figure(self):
        if self.main_window is None:
            return
        self.main_window.fig_view.draw()

    def refresh_controls(self):
        if self.main_window is None:
            return
        self.main_window.refresh_controls()

    def refresh_control(self, name):
        if self.main_window is None:
            return
        self.main_window.refresh_control(name)

    def set_control_min_max(self, name, min_value, max_value):
        if self.main_window is None:
            return
        self.main_window.set_control_min_max(name, min_value, max_value)

    def set_control_title(self, name, new_title):
        if self.main_window is None:
            return
        self.main_window.set_control_title(name, new_title)

    def set_and_refresh_control(self, control_name, value):
        self.state[control_name] = value
        self.refresh_control(control_name)

    def set_control_value(self, name, value):
        self.state[name] = value
        self.control_changed(name, value)
        self.refresh()

    def set_control_value_internal(self, name, value):
        """
        Set a new value for control without firing the control_changed event
            Useful when the value is paired with an internal variable and because the slider rounds value some
            precision is lost (see self.orientation_key in view_localization.py for an example)
        :param name:
        :param value:
        :return: nothing
        """
        # print('%s changed to %s (internally)' % (name, value))
        self.state[name] = value

    def set_control_value_if_not_present(self, name, value):
        """
        Set control value only if name is not already in state
        Useful for Label controls, where the title is the initial value but later can be changed
        :param name:
        :param value:
        :return:
        """
        if name not in self.state:
            self.set_control_value(name, value)

    def control_changed(self, name, value):
        """
        Abstract method, used to trigger actions after a control changed
        :param name:
        :param value:
        :return:
        """
        pass

    def get_control_by_name(self, name):
        if self.main_window is None:
            return None
        return self.main_window.get_control_by_name(name)

    def get_control_value(self, name):
        if name in self.state:
            return self.state[name]
        else:
            value = self.calculated_value(name)
            return value if value is not None else 0

    def control_has_value(self, name):
        return name in self.state

    def controls_def(self):
        # abstract method
        return []

    def initialize(self):
        # abstract method, use to initialize stuff
        pass

    def get_file_name(self, title='Open', file_filter='*.*', directory=''):
        """
        Returns a valid file name using the standard dialog
        :param title:        window title
        :param file_filter:  show file that comply with this filter only
        :param directory:    initial directory ('' means current one)
        :return: a file name (None means no file was chosen)
        """
        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(self.main_window, title, filter=file_filter,
                                                             directory=directory)
        return file_name if file_name != '' else None

    def get_file_name_to_save(self, title='Save', file_filter='*.*', directory=''):
        """
        Returns file name to be saved (can be a new one or an existing one
        :param title:
        :param file_filter:
        :param directory:
        :return:
        """
        file_name, _ = QtWidgets.QFileDialog.getSaveFileName(self.main_window, title, filter=file_filter,
                                                             directory=directory)
        return file_name if file_name != '' else None

    def get_progress_bar(self):
        """
        Returns a progress bar to be displayed in the status bar
        :return:
        """
        # if main_windows is not set yet just return an empty one in order not to crash,
        # this it not a problem because main_window will be set later
        return self.main_window.progress_bar if self.main_window is not None else GeneralProgressBar()

    def title(self):
        # abstract method
        return None   # if None then the config title will be used

    def view_size(self):
        return 10, 10

    def axes_limits(self):
        # check for empty is necessary for on_click event
        return self.box_size.size() if not self.box_size.is_empty else None

    def text_position(self):
        if not self.box_size.is_empty:
            min_x, _, _, max_y = self.box_size.size()
            return min_x, max_y
        return self.main_window.win_config.get('text_position', None) if self.main_window is not None else None

    def show_figure_text(self, ax, text_values):
        position = self.text_position()
        if position is None:
            return
        show_text_values(ax, text_values, position[0], position[1])

    def show_status_bar_msg(self, msg):
        if self.main_window is None:
            print('no main window defined yet')
            return
        self.main_window.show_status_bar_msg(msg)

    def update_view(self, ax):
        # abstract method
        pass

    # Zoom management
    def zoom_active(self):
        return self.state.get(self.zoom_key, False)

    def get_zoom_actions(self, event):
        return [['Zoom in', functools.partial(self.set_zoom_in, event)], ['Zoom out', self.set_zoom_out]]

    def set_zoom_out(self):
        self.state[self.zoom_key] = False
        self.refresh()

    def set_zoom_in(self, event):
        self.state[self.zoom_key] = True
        self.zoom_center = [event.xdata, event.ydata]
        self.refresh()

    def toggle_zoom(self):
        self.state[self.zoom_key] = not self.state[self.zoom_key]
        if not self.state[self.zoom_key]:
            self.refresh()

    def set_zoom_center(self, event):
        if self.zoom_active():
            self.zoom_center = [event.xdata, event.ydata]
            self.refresh()

    def apply_zoom(self):
        if self.zoom_active() and self.zoom_center is not None:
            self.box_size.reset()
            self.box_size.add_points([[self.zoom_center[0] - self.zoom_radius, self.zoom_center[1] - self.zoom_radius],
                                      [self.zoom_center[0] + self.zoom_radius, self.zoom_center[1] + self.zoom_radius]])

    # figure sizing
    def resize_figure(self, ax, points, fixed_points=(), with_reset=True, inc=1.1):
        """
        Recalculate the figure size with the given points and resize it
        :param ax:
        :param points: :type list of [x, y]
        :param fixed_points: list of point to add in addition to points
        :param with_reset:
        :param inc:
        :return:
        """
        if with_reset:
            self.box_size.reset()
        self.box_size.add_points(points)
        self.box_size.add_points(fixed_points)
        self.box_size.set_bounds(ax, inc)

    def on_mouse_click(self, event, ax, parent):
        # abstract method
        # print('%s click: button=%d, x=%d, y=%d, xdata=%f, ydata=%f' %
        #      ('double' if event.dblclick else 'single', event.button, event.x, event.y, event.xdata, event.ydata))
        if event.button == QTAux.MouseButton.Right:
            return self.on_right_click(event, ax, parent)
        else:
            return self.onclick(event, ax, parent)

    def onclick(self, event, ax, parent):
        # abstract method
        # print('%s click: button=%d, x=%d, y=%d, xdata=%f, ydata=%f' %
        #      ('double' if event.dblclick else 'single', event.button, event.x, event.y, event.xdata, event.ydata))
        pass

    def on_right_click(self, event, ax, parent):
        # abstract method
        # print('%s click: button=%d, x=%d, y=%d, xdata=%f, ydata=%f' %
        #      ('double' if event.dblclick else 'single', event.button, event.x, event.y, event.xdata, event.ydata))
        pass

    def on_mouse_move(self, event, ax):
        # abstract method
        # print('%s click: button=%d, x=%d, y=%d, xdata=%f, ydata=%f' %
        #      ('double' if event.dblclick else 'single', event.button, event.x, event.y, event.xdata, event.ydata))
        return False

    def calculated_value(self, name):
        # abstract method
        return None

    def save_cycle(self):
        """
        Save the info of a given cycle
        Note: it is directly called from an event in the UI (as defined in a yaml config file)
        :return:
        """
        file_name = 'cycle'
        r = rc.Record(file_name)
        to_store = self.get_info_to_save()
        r.write_group('cycle', to_store)
        print('File %s saved' % r.get_full_file_name())

    def get_info_to_save(self):
        # abstract method
        return {}

    # animation methods
    def get_data_provider(self, interval=100, min_x=0.0, max_x=10.0, data_provider=None):
        """
        Abstract method
        :param interval: time at which show new data (in milliseconds)
        :param max_x:    max value in the x-axis, values greater than this will cause the graph to scroll left
        :param min_x:    start value for the x-axis
        :param data_provider:  subclass of RealTimeDataProvider (can be None)
        :return: interval (milliseconds), number of points to show, x axis bounds (ex: [-1, 100]), subclass of
                 RealTimeDataProvider (can be None)
        """
        dt            = interval/1000  # seconds
        max_points    = int(max_x/dt) + 1
        return interval, max_points, (min_x, max_x), data_provider

    def anim_is_running(self):
        return self.main_window.fig_view.anim_is_running

    def stop_animation(self):
        self.main_window.fig_view.stop_animation()

    def start_animation(self):
        self.main_window.fig_view.start_animation()


class PropertiesHost(HostModel):
    """
    HostModel specialized for editing a set of properties (defined as a dict)
    """

    def __init__(self, dialog_name, properties):
        self.dialog_name   = dialog_name  # config name for the dialog
        self.properties    = properties
        self.changed       = {}
        self.was_confirmed = False

        self.status_msg = 'Show %s' % self.dialog_name
        initial_values = self.properties.copy()
        super(PropertiesHost, self).__init__(initial_values=initial_values)
        self.dialog = Dialog(self.dialog_name, self)

    def set_properties(self, properties):
        self.properties = properties

    def show(self):
        self.was_confirmed = False
        self.update_state()
        self.dialog.show()
        return self.changed

    def button_confirm(self):
        self.changed = self.get_changed()
        self.dialog.close()
        self.was_confirmed = True
        self.after_confirm(self.changed)

    def button_cancel(self):
        self.dialog.close()

    def after_confirm(self, changed):
        # abstract method
        pass

    def update_properties(self, changed):
        if changed is None:
            return
        self.properties.update(changed)

    def update_state(self):
        for k, v in self.properties.items():
            self.set_control_value(k, v)
            self.dialog.refresh_control(k)

    def get_changed(self):
        changed = {}
        for k, v in self.properties.items():
            if k not in self.state or v == self.state[k]:
                continue
            changed[k] = self.state[k]
        return changed


class TestHost(HostModel):
    """
    HostModel specialized for showing a set of tests stored in a file.
        The file name and the test name must be defined in the .yaml config file
    """

    def __init__(self, initial_values):
        # keys
        self.test_key   = 'test'

        # test cases
        self.all_cases = []
        self.output    = None

        super(TestHost, self).__init__(initial_values=initial_values)

    def initialize(self):
        test_file_name = self.main_window.win_config.get('test_file_name', '')
        test_name      = self.main_window.win_config.get('test_name', '')
        test_file      = yaml.get_yaml_file(test_file_name)
        test           = yaml.get_record(test_file['general'], test_name, 'tests', 'test', alternative_key_name='call')
        self.all_cases = test['cases']
        self.set_control_min_max(self.test_key, 0, len(self.all_cases)-1)
        self.set_current_case()

    def control_changed(self, name, _):
        if name == self.test_key:
            self.set_current_case()

    def set_current_case(self):
        case = self.get_current_case()
        if case is None:
            self.set_no_case()
            return

        self.box_size.reset()  # to avoid resizing window when click
        current_case = case['case']
        self.output  = current_case.get('output', None)
        desc = self.set_case(current_case['input'],  self.output, current_case.get('desc', ''))
        self.show_status_bar_msg(desc)
        self.refresh_controls()

    def get_current_case(self):
        if not self.all_cases:
            return None
        return self.all_cases[int(self.state[self.test_key])]

    def get_current_description(self, test_description):
        return test_description

    def set_case(self, case_input, output, description):
        """Abstract method for setting particular data"""
        return description

    def set_no_case(self):
        """Abstract method for setting null data when no case is ready"""
        pass


# Layout definition
def get_win_config(config_file_name, window_key='window'):
    config1 = yaml.get_yaml_file(config_file_name)
    win_config = config1.get(window_key, {})
    if win_config == {}:
        raise Exception('"window" keyword not present in config file %s' % config_file_name)
    return win_config


def get_title(win_config, provider, key='title'):
    title = provider.title()
    if title is None:
        title = win_config.get(key, 'Title')
    return title


def set_grid_layout(father_layout, subtype, layout_config, controls, window, row_col=None):
    if subtype == 'vertical':
        layout = QtWidgets.QVBoxLayout()
    elif subtype == 'horizontal':
        layout = QtWidgets.QHBoxLayout()
    else:
        raise Exception('Subtype %s not implemented for %s' % (subtype, type))
    add_controls_to_window(layout, layout_config, controls, window)
    if not row_col:
        father_layout.addLayout(layout)
    else:
        father_layout.addLayout(layout, row_col[0], row_col[1])
    return layout


def set_layout(father_layout, layout_config, controls, window, row_col=None):
    fig_view = None
    if father_layout is None or not layout_config:
        return fig_view
    for sub_layout_config1 in layout_config:
        sub_layout_config = sub_layout_config1['item']
        sub_layout, fig_view_sub = set_sub_layout(father_layout, sub_layout_config, controls, window, row_col)
        if fig_view_sub is not None:
            fig_view = fig_view_sub
        fig_view1 = set_layout(sub_layout, sub_layout_config.get('layout', []), controls, window)
        if fig_view1 is not None:
            fig_view = fig_view1
    return fig_view


def set_sub_layout(father_layout, layout_config, controls, window, row_col=None):
    fig_view    = None
    layout_type = layout_config['type']
    subtype     = layout_config.get('subtype', None)
    if layout_type == 'grid':
        layout = set_grid_layout(father_layout, subtype, layout_config, controls, window, row_col=row_col)
    elif layout_type == 'figure':
        fig_view = set_figure_layout(father_layout, layout_config, window)
        layout   = None
    else:
        print('WARNING: layout type "%s" not implemented' % layout_type)
        layout = None
    return layout, fig_view


def set_figure_layout(father_layout, layout_config, window, key_axes='axes_limits', subtype_key='subtype'):
    axes_limits = layout_config.get(key_axes, None)
    subtype     = layout_config.get(subtype_key, None)
    size        = window.provider.view_size() if window.provider is not None else (1, 0)
    fig_view    = FigureView(window, size=size, axes_limits=axes_limits, subtype=subtype)
    father_layout.addWidget(fig_view)
    return fig_view


def create_menu_bar(config, main_window, provider, key='menu_bar'):
    if key not in config:
        return None
    menu_bar = QtWidgets.QMenuBar(parent=main_window)
    menu_bar_config = config[key]
    for main_item1 in menu_bar_config:
        main_item = main_item1['item']
        main_menu = QtWidgets.QMenu(main_item['title'], main_window)
        menu_bar.addMenu(main_menu)
        if 'items' in main_item:
            for sub_item1 in main_item['items']:
                sub_item = sub_item1['item']
                if sub_item.get('is_separator', False):
                    main_menu.addSeparator()
                    continue

                title  = sub_item['title']
                action = QtWidgets.QAction(title, main_window)
                action.triggered.connect(functools.partial(exec_action, provider,
                                                           function_name=sub_item.get('action', None)))
                # action = QTAux.MenuItem('1', sub_item['title'], self.provider, sub_item.get('action', None),
                #                        None, self)
                main_menu.addAction(action)
    return menu_bar


def create_toolbar(config, main_window, provider, controls, key='toolbar', icon_key='icon', tooltip_key='tooltip',
                   combo_key='Combo', check_key='Check', type_key='type', label_key='Label', action_key='Action'):
    if key not in config:
        return None
    toolbar        = QtWidgets.QToolBar()
    toolbar_config = config[key]
    for item1 in toolbar_config:
        item      = item1['item']
        title     = item.get('title', '')
        name      = item.get('name', 'NoName')
        item_type = item.get(type_key, 'Action')
        tooltip   = item.get(tooltip_key, None)

        if item.get('is_separator', False):
            toolbar.addSeparator()
        elif item_type == combo_key:
            controls.append(QTAux.Combo(name, title, provider, None, toolbar, item))
        elif item_type == check_key:
            controls.append(QTAux.CheckButton(name, title, provider, None, toolbar, tooltip))
        elif item_type == label_key:
            QTAux.Label(name, title, provider, None, toolbar, align_left=False)
        elif item_type == action_key:
            provider.set_control_value(name, title)
            action = QTAux.Action(name, title, provider, main_window, item, tooltip=tooltip)
            action.qt_action.triggered.connect(functools.partial(exec_action, provider,
                                                                 function_name=item.get('action', None)))
            toolbar.addAction(action.qt_action)
            controls.append(action)
        else:
            raise Exception('%s is not implemented in Toolbar' % item_type)

    return toolbar


def create_status_bar(win_config, main_window, status_key='status_bar'):
    if not win_config.get(status_key, True):  # create a statusbar unless is explicitly forbidden
        return None
    statusbar = QtWidgets.QStatusBar(main_window)
    main_window.setStatusBar(statusbar)
    return statusbar


class GeneralProgressBar:
    """
    Handle a ProgressBar to be used inside any widget, main use is to put it in a StatusBar
    Note: if widget is None all methods must work anyway, in order to free all calling usage to check whether
          ProgressBar is present or not
    """

    def __init__(self, widget=None, stretch=0, visible=False):
        """
        init
        :param widget:  widget where the progress bar will show. due to some initialization timing sometime widget can
                        be None
        :param stretch: amount of extra space to be used inside widget. 0 means original size, usually 1 is enough to
                        guarantee it will occupy the full widget
        :param visible:
        """
        self.parent = widget
        if self.parent is None:
            return
        self.progress_bar = QtWidgets.QProgressBar(parent=self.parent)
        self.parent.addWidget(self.progress_bar, stretch=stretch)
        self.progress_bar.setVisible(visible)

    def get_maximum(self):
        if self.parent is None:
            return
        return self.progress_bar.maximum()

    def get_value(self):
        if self.parent is None:
            return
        return self.progress_bar.value()

    def reset(self, text=None):
        if self.parent is None:
            return
        self.progress_bar.reset()
        self.progress_bar.setVisible(False)
        if text is not None and self.parent is not None:
            self.parent.showMessage(text)

    def set_max(self, max_value):
        if self.parent is None:
            return
        self.progress_bar.setMaximum(max_value)
        self.progress_bar.setVisible(True)

    def set_value(self, value):
        if self.parent is None:
            return
        self.progress_bar.setValue(value)

    def add_increment(self, increment):
        if self.parent is None:
            return
        self.progress_bar.setValue(self.progress_bar.value()+increment)


# Controls definition
def add_controls_to_window(layout, layout_config, controls, window):
    controls_config = layout_config.get('controls', [])
    controls.extend(def_controls(controls_config, window.provider, layout))


def def_controls(controls_definition, provider, layout):
    controls = []
    for control in controls_definition:
        valid, c = def_control(control, provider, layout)
        if valid:
            controls.append(c)
    return controls


def def_control(control1, provider, layout, main_key='control'):
    control = control1[main_key]
    c_name  = control['name']
    e_name  = external_name(control)
    c_type  = control['type']
    tooltip = control.get('tooltip', None)
    action  = action_string(control)
    value   = get_control_initial_value(c_name, control, provider)
    # print('%s initial value: %s' % (c_name, value))
    if value is not None:
        # value must be set before control creation, so it will appear the first time the control is displayed
        provider.set_control_value(c_name, value)

    if c_type == 'Slider':
        result = True
        qt_control = def_slider(c_name, e_name, provider, layout, control['parms'], action, tooltip=tooltip)
    elif c_type == 'Combo':
        result = True
        qt_control = QTAux.Combo(c_name, e_name, provider, action, layout, control, tooltip=tooltip)
    elif c_type == 'EnumCombo':
        result     = True
        enum       = eval(control['enum'])
        qt_control = QTAux.EnumCombo(c_name, e_name, provider, enum, action, layout)
        # qt_control = def_enum_combo(c_name, e_name, provider, layout, action, control)
    elif c_type == 'Button':
        result = True
        qt_control = def_button(c_name, e_name, provider, layout, action, control, tooltip=tooltip)
    elif c_type == 'Constant':
        result = False
        qt_control = def_constant(c_name, e_name, provider, layout, action)
    elif c_type == 'Check':
        result = True
        qt_control = QTAux.CheckButton(c_name, e_name, provider, action, layout, tooltip=tooltip)
    elif c_type == 'Text':
        result = True
        qt_control = QTAux.Label(c_name, e_name, provider, action, layout)
    elif c_type == 'EditText':
        result = True
        qt_control = QTAux.EditText(c_name, e_name, provider, action, layout, tooltip=tooltip)
    elif c_type == 'EditNumber':
        result = True
        qt_control = QTAux.EditNumber(c_name, e_name, provider, action, layout)
    elif c_type == 'EditNumberSpin':
        result = True
        qt_control = QTAux.EditNumberSpin(c_name, e_name, provider, action, layout, control)
    elif c_type == 'ProgressBar':
        result = True
        qt_control = QTAux.ProgressBar(c_name, provider, action, layout)
    else:
        raise Exception('Control type %s not implemented' % c_type)

    qt_control.refresh_others = control.get('refresh_others', False)

    return result, qt_control


def get_control_initial_value(c_name, control, provider, key_value='value'):
    if key_value in control:
        return control[key_value]
    elif provider.control_has_value(c_name):
        return provider.get_control_value(c_name)
    else:
        return None


def def_slider(c_name, e_name, provider, layout, parms, action, tooltip=None):
    [min_e, max_e, scale_e] = parms
    min1   = get_def_value(min_e, provider)
    max1   = get_def_value(max_e, provider)
    scale1 = get_def_value(scale_e, provider)
    return QTAux.Slider(c_name, e_name, provider, min1, max1, action, layout, tooltip=tooltip, scale=scale1)


def def_button(c_name, e_name, provider, layout, action, control, tooltip=None, def_width=30, def_length=100):
    if action != '':
        width  = control.get('width', def_width)
        length = control.get('length', def_length)
        return QTAux.Button(c_name, e_name, provider, action, layout, tooltip=tooltip, width=width, length=length)
    else:
        raise Exception('Invalid button definition, no action defined')


def def_constant(c_name, e_name, provider, layout, action):
    # Only reason to have a constant on window is to exec an action
    if action != '':
        QTAux.Constant(c_name, e_name, provider, action, layout)
    else:
        raise Exception('Invalid constant definition, no action defined')


def external_name(item, title_key='title', desc_key='desc', name_key='name'):
    if title_key in item:
        k = title_key
    elif desc_key in item:
        k = desc_key
    else:
        k = name_key
    return item[k]


def get_with_default(values, key, default):
    if key in values:
        return values[key]
    else:
        return default


def get_def_value(v, provider):
    # Note: provider is necessary so eval can use it
    string = QTAux.string_to_eval(v)
    return eval(string) if string is not None else v


def action_string(control_def, action_key='action'):
    return control_def.get(action_key, '')


def exec_action(provider, function_name=None):
    # Note: provider is necessary so eval can use it
    if function_name is None:
        return
    full_function_name = 'provider.%s()' % function_name
    eval(full_function_name)


# Visualizations
def show_text_values(ax, values, x, y, boxstyle='round', facecolor='wheat', alpha=0.5, vertical_alignment='top'):
    """
    Display a list of values (name=value) in a text box
    :param ax:
    :param values: :type list of [nane, value]
    :param x:
    :param y:
    :param boxstyle:
    :param facecolor:
    :param alpha:
    :param vertical_alignment:
    :return:
    """
    if not values:
        return
    text_string = ''
    next1       = ''
    for [name, value] in values:
        text_string += r'%s$%s=%s$' % (next1, name, value)
        next1 = '\n'
    props = dict(boxstyle=boxstyle, facecolor=facecolor, alpha=alpha)
    ax.text(x, y, text_string, verticalalignment=vertical_alignment, bbox=props)


def get_arg_value(i, default):
    return default if len(sys.argv) < i + 1 else sys.argv[i]



#!/usr/bin/env python

import sys
from PyQt5 import QtCore, QtWidgets, QtGui
from enum import IntEnum


def def_app(style='Plastique'):
    app = QtWidgets.QApplication(sys.argv)
    QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create(style))
    return app


# You need to set up a signal slot mechanism, to
# send data to your GUI in a thread-safe way.
# Believe me, if you don't do this right, things
# go very wrong ...
class Communicate(QtCore.QObject):
    data_signal = QtCore.pyqtSignal(list)


def set_button(name, width, height, layout, action):
    button = QtWidgets.QPushButton(name)
    set_custom_size(button, width, height)
    button.clicked.connect(action)
    layout.addWidget(button)
    return button


def set_label(name, layout, height=None, width=None):
    label = QtWidgets.QLabel(name)
    label.setAlignment(QtCore.Qt.AlignCenter)
    if height is not None:
        label.setFixedHeight(height)
    if width is not None:
        label.setFixedWidth(height)
    layout.addWidget(label)
    return label


def set_window(window, title, size):
    [new_x, new_y, width, height] = size
    window.setGeometry(new_x, new_y, width, height)
    window.setWindowTitle(title)


def set_custom_size(x, width, height):
    size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
    size_policy.setHorizontalStretch(0)
    size_policy.setVerticalStretch(0)
    size_policy.setHeightForWidth(x.sizePolicy().hasHeightForWidth())
    x.setSizePolicy(size_policy)
    x.setMinimumSize(QtCore.QSize(width, height))
    x.setMaximumSize(QtCore.QSize(width, height))


def set_progress_bar(host):
    return QtWidgets.QProgressBar(host)


class ScreenControl(object):
    """
    Abstract class for any kind of input control like Sliders and Combos
    """

    def __init__(self, name, title, bound, action, layout, tooltip=None, label_height=20):
        # bound is the object responsible to keep the value, must implement:
        #     get_control_value
        #     set_control_value
        self.name    = name
        self.ename   = title
        self.bounded = bound
        self.action  = action

        # print('control:%s bound:%s, action:%s' % (name, bound, action))

        self.bounded_name = 'self.bounded.'
 
        is_widget, widget = self.get_widget()
        if is_widget and layout is not None:
            title = self.title()
            if title != '':
                self.label = set_label(title, layout, height=label_height)  # must go after slider (see title())
            layout.addWidget(widget)
            if tooltip is not None:
                widget.setToolTip(tooltip)

        self.refresh_others = False
        self.refresh()

    def current_value(self):
        return self.bounded.get_control_value(self.name)

    def title(self):
        return self.ename

    def changed(self):
        # print('%s changed to %s (bounded:%s)' % (self.name, self.value(), self.bounded))
        self.bounded.set_control_value(self.name, self.value())
        self.exec_action()

    def exec_action(self):
        self.exec_bounded_method(self.action)

    def exec_bounded_method(self, method_name):
        if method_name is None or method_name == '':
            return
        return getattr(self.bounded, method_name)()

    def get_widget(self):
        # abstract method
        return False, None

    def value(self):
        # abstract method
        return 0

    def refresh(self):
        # abstract method
        pass

    def set_ename(self, new_ename):
        self.ename = new_ename


class EnumCombo(ScreenControl):
    def __init__(self, name, title, bound, enum, action, layout, tooltip=None):
    
        self.enum  = enum
        self.combo = QtWidgets.QComboBox(None)
        for member in self.enum:
            self.combo.addItem(member.name)

        self.combo.currentIndexChanged.connect(self.changed_combo)

        super(EnumCombo, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

    def current_value(self):
        e = self.bounded.get_control_value(self.name)
        return e  # e.value

    def value(self):
        value = self.enum(self.combo.currentIndex())
        return value

    def refresh(self):
        self.combo.setCurrentIndex(self.current_value())

    def changed_combo(self, i):
        self.bounded.set_control_value(self.name, self.enum(i))
        self.exec_action()

    def get_widget(self):
        return True, self.combo


class Combo(ScreenControl):
    def __init__(self, name, title, bound, action, layout, control_def, tooltip=None, values_key='values'):
        self.combo = QtWidgets.QComboBox(None)

        self.combo.currentTextChanged.connect(self.changed)
        super(Combo, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

        original_values         = []
        self.load_values_action = None
        if values_key in control_def:
            values1 = control_def[values_key]
            if isinstance(values1, list):
                original_values = values1
            else:
                load_function = string_to_eval(values1)
                if load_function is not None:
                    self.load_values_action = load_function
        self.set_values(original_values)  # must go after definition so self.bounded_name exists

    def set_values(self, values):
        self.combo.clear()
        values1 = self.exec_bounded_method(self.load_values_action) if self.load_values_action is not None else values
        self.combo.addItems(values1)

    def value(self):
        return self.combo.currentText()

    def get_widget(self):
        return True, self.combo

    def refresh(self):
        # print('current value: %s of %s' % (self.current_value(), self.name))
        self.combo.setCurrentText(self.current_value())


class Slider(ScreenControl):
    def __init__(self, name, title, bound, min_value, max_value, action, layout, tooltip=None, scale=1):

        # Widget must be created before calling super
        self.slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.set_min_max(min_value, max_value)
        self.vfactor = scale
        self.action  = action
        
        self.slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        interval = (max_value - min_value)/10
        self.slider.setTickInterval(interval)

        self.slider.valueChanged.connect(self.changed)

        super(Slider, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

    def title(self):
        return '%s(%s)' % (self.ename, self.value())

    def changed(self):
        # print('set %s' % self.name)
        self.bounded.set_control_value(self.name, self.value())
        self.label.setText(self.title())
        self.exec_action()

    def get_widget(self):
        return True, self.slider

    def value(self):
        value = float(self.slider.value())/self.vfactor  # integer to float scaling (slider only accept integers)
        return value

    def refresh(self):
        self.slider.setValue(self.current_value()*self.vfactor)

    def set_min_max(self, min_value, max_value):
        # print('set min:%s max:%s' % (min_value, max_value))
        self.slider.setMinimum(min_value)
        self.slider.setMaximum(max_value)


class Button(ScreenControl):
    def __init__(self, name, title, bound, action, layout, tooltip=None, width=50, length=100):

        # Widget must be created before calling super
        self.button = set_button(title, length, width, layout, self.changed)
        super(Button, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

    def title(self):
        return ''  # button don't have title label

    def changed(self):
        # print '%s clicked (%s)' %(self.name, self.event)
        self.exec_action()

    def get_widget(self):
        return True, self.button

    def set_ename(self, new_ename):
        self.ename = new_ename
        self.button.setText(self.ename)


class CheckButton(ScreenControl):
    def __init__(self, name, title, bound, action, layout, tooltip=None):

        # Widget must be created before calling super
        self.button = QtWidgets.QCheckBox(title)
        super(CheckButton, self).__init__(name, title, bound, action, layout, tooltip=tooltip)
        self.button.stateChanged.connect(self.changed)
        self.button.setChecked(self.current_value())

    def title(self):
        return ''  # button don't have title label

    def value(self):
        return self.button.isChecked()

    def get_widget(self):
        return True, self.button

    def refresh(self):
        self.button.setChecked(self.current_value())


class EditText(ScreenControl):
    def __init__(self, name, title, bound, action, layout, tooltip=None):
        # Widget must be created before calling super
        self.edit_text = QtWidgets.QLineEdit()
        self.edit_text.textChanged.connect(self.changed)
        super(EditText, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

    def get_widget(self):
        return True, self.edit_text

    def value(self):
        value = self.edit_text.text()
        return value

    def refresh(self):
        self.edit_text.setText(str(self.current_value()))


class EditNumber(ScreenControl):
    def __init__(self, name, title, bound, action, layout, tooltip=None):
        # Widget must be created before calling super
        self.edit_text = QtWidgets.QLineEdit()
        self.edit_text.setValidator(QtGui.QDoubleValidator())
        self.edit_text.textChanged.connect(self.changed)
        super(EditNumber, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

    def get_widget(self):
        return True, self.edit_text

    def value(self):
        value = float(self.edit_text.text())
        return value

    def refresh(self):
        self.edit_text.setText(str(self.current_value()))


class EditNumberSpin(ScreenControl):
    def __init__(self, name, title, bound, action, layout, control_def, tooltip=None, parms_def='parms',
                 step_def='step', type_def='type', min_def='minimum', max_def='maximum'):
        # Widget must be created before calling super
        parms = control_def.get(parms_def, {})
        is_integer = parms.get(type_def, 'float') == 'integer'
        self.edit_spin = QtWidgets.QSpinBox(None) if is_integer else QtWidgets.QDoubleSpinBox(None)
        if step_def in parms:
            self.edit_spin.setSingleStep(parms[step_def])
        if min_def in parms:
            self.edit_spin.setMinimum(parms[min_def])
        if max_def in parms:
            self.edit_spin.setMaximum(parms[max_def])
        self.edit_spin.valueChanged.connect(self.changed)
        super(EditNumberSpin, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

    def get_widget(self):
        return True, self.edit_spin

    def value(self):
        value = float(self.edit_spin.value())
        return value

    def refresh(self):
        self.edit_spin.setValue(self.current_value())

    def set_min_max(self, min_value, max_value):
        # print('set min:%s max:%s' % (min_value, max_value))
        self.edit_spin.setMinimum(min_value)
        self.edit_spin.setMaximum(max_value)


class MenuItem(ScreenControl):
    def __init__(self, name, title, bound, action, layout, main_window, tooltip=None):
        # Widget must be created before calling super
        self.menu_item = QtWidgets.QAction(title, main_window)
        if action is not None:
            self.menu_item.triggered.connect(self.changed)
        super(MenuItem, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

    def get_widget(self):
        return False, self.menu_item


class Action(ScreenControl):
    """
    Wrapper for a QtAction, useful
    """
    def __init__(self, name, title, provider, main_window, properties, tooltip=None, icon_key='icon'):
        self.name = name
        if icon_key in properties:
            std_icon_name = getattr(QtWidgets.QStyle, properties[icon_key])
            icon = main_window.style().standardIcon(std_icon_name)
            self.qt_action = QtWidgets.QAction(icon, title, main_window)
        else:
            self.qt_action = QtWidgets.QAction(title, main_window)

        if tooltip is not None:
            self.qt_action.setToolTip(tooltip)
        super(Action, self).__init__(name, title, provider, None, None, tooltip=None)

    def value(self):
        return self.ename

    def refresh(self):
        # print('refresh title:%s' % self.current_value())
        self.qt_action.setText(self.current_value())

    def set_ename(self, new_ename):
        self.ename = new_ename
        self.qt_action.setText(self.ename)


class Label(ScreenControl):
    """
    Read only text (also known as Label)
    """
    def __init__(self, name, title, bound, action, layout, tooltip=None, align_left=True):

        # Widget must be created before calling super
        self.text = QtWidgets.QLabel(title)
        if align_left:
            self.text.setAlignment(QtCore.Qt.AlignLeft)
        bound.set_control_value_if_not_present(name, title)
        super(Label, self).__init__(name, title, bound, action, layout, tooltip=tooltip)

    def title(self):
        # returns '' to avoid creating another Label
        return ''

    def get_widget(self):
        return True, self.text

    def refresh(self):
        # print('current value:%s' % self.current_value())
        self.text.setText('%s' % self.current_value())


class ProgressBar(ScreenControl):
    def __init__(self, name, bound, action, layout, tooltip=None):

        # Widget must be created before calling super
        self.progress_bar = QtWidgets.QProgressBar(None)
        super(ProgressBar, self).__init__(name, '', bound, action, layout, tooltip=tooltip)

    def get_widget(self):
        return True, self.progress_bar

    def get_maximum(self):
        return self.progress_bar.maximum()

    def get_value(self):
        return self.progress_bar.value()

    def set_max(self, max_value):
        self.progress_bar.setMaximum(max_value)
        self.progress_bar.setVisible(True)

    def set_value(self, value):
        self.progress_bar.setValue(value)

    def add_increment(self, increment):
        self.progress_bar.setValue(self.get_value()+increment)

    def set_visible(self, value):
        self.progress_bar.setVisible(value)

    def reset(self):
        self.progress_bar.reset()


class Constant(ScreenControl):
    def __init__(self, name, title, bound, action, layout, tooltip=None):
        # it doesn't show anything, it is used for executing actions at initialization time
        super(Constant, self).__init__(name, title, bound, action, layout, tooltip=tooltip)
        self.cte_value = self.current_value()
        self.exec_action()

    def value(self):
        return self.cte_value

    def get_widget(self):
        # Don't add as widget
        return False, 0


class Menu:
    """
    Create a contextual menu to be showed via .popup()
    """
    def __init__(self, parent, title='Menu', actions=()):
        """
        :param parent:
        :param title:
        :param actions: list of [action_name, function_to_call]
        """
        self.menu = QtWidgets.QMenu(title, parent=parent)
        for [name, event] in actions:
            action = self.menu.addAction(name)
            action.triggered.connect(event)

    def popup(self):
        cursor = QtGui.QCursor()
        self.menu.popup(cursor.pos())


def string_to_eval(v):
    """
    Check if v is a string to the type '= ...', if not returns false,
    if true returns the string without the = (so it can be evaluated directly)
    :param v:
    :return:
    """
    return v[1:] if isinstance(v, str) and v[0] == '=' else None


class MouseButton(IntEnum):
    Left  = 1
    Right = 3



from collections import deque


class SignalHistory(object):
    """
    Store the last N values of a Signal
    """

    def __init__(self, length=10):
        self.length = length + 1
        self.values = None  # just to avoid warnings
        self.reset()

    def reset(self):
        self.values = deque(maxlen=self.length)

    def get_len(self):
        return self.length

    def set_len(self, length):
        """
        Changes the length of the queue
        Note: deletes the current values
        :param length:
        :return:
        """
        self.length = length + 1
        self.reset()

    def append(self, value):
        self.values.append(value)

    def update(self, value):
        """
        Change the last value
        :param value:
        :return:
        """
        # print 'before:%s v:%s' %(self.values, value)
        self.values.pop()
        self.values.append(value)
        # print 'after:%s' %(self.values)

    def load(self, values):
        """
        Bulk store values. Used to load from telemetry
        :param values:
        :return:
        """
        self.reset()
        for value in values:
            self.values.append(value)

    def is_full(self):
        return len(self.values) >= self.length

    def get_items_in_lifo_order(self):
        """
        Returns item in Last In - First Out order
        :return:
        """
        for i in range(len(self.values)-1, -1, -1):
            yield self.values[i]

    def get_items_in_fifo_order(self):
        """
        Returns item in First In - First Out order
        :return:
        """
        for v in self.values:
            yield v

    # aggregates
    def get_aggregate(self, aggregate_type):
        """
        Returns one of the implemented aggregates type, useful to test many kind of filters
        :param aggregate_type:
        :return:
        """
        if aggregate_type == 'last':
            return self.last()
        elif aggregate_type == 'average':
            return self.average()
        elif aggregate_type == 'min':
            return self.min()
        elif aggregate_type == 'max':
            return self.max()
        elif aggregate_type == 'sum':
            return self.sum()
        else:
            raise Exception('%s aggregate type not implemented' % aggregate_type)

    def last(self):
        return self.values[-1]

    def average(self):
        sum_v   = self.sum()
        count_v = len(self.values)
        if count_v > 0:
            # print 's:%s c:%s' %(sum_v, count_v)
            return float(sum_v)/count_v
        else:
            return 0

    def min(self):
        return min(self.values)

    def max(self):
        return max(self.values)

    def sum(self):
        # To Do: optimize to have the sum pre-calculated in every update
        return sum(self.values)

    def weighted_sum(self, weights, lifo_order=True):
        function = self.get_items_in_lifo_order if lifo_order else self.get_items_in_fifo_order
        total = 0.0
        for i, value in enumerate(function()):
            total += weights[i]*value
        return total
    # end aggregates

    def local_optimum_points(self):
        # returns the points where there are a local min or max
        dif = 0.1
        points = []
        for i in range(2, len(self.values)):
            d1 = self.values[i-1] - self.values[i-2]
            d2 = self.values[i]   - self.values[i-1]
            if d2 < -dif:
                if d1 > dif:
                    # local max
                    points.append((i, self.values[i-1], 1))  # 1 means max
            else:
                if d1 < -dif:
                    # local min
                    points.append((i, self.values[i-1], -1))  # -1 means min
        return points

    def changed_derivative(self):
        """
        Returns True if values changed derivative in the period
        :return: :type boolean
        """
        for i in range(2, len(self.values)):
            d1 = self.values[i-1] - self.values[i-2]
            d2 = self.values[i]   - self.values[i-1]
            if abs(d2-d1) > 0.1:  # not np.sign(d1) == np.sign(d2):
                # print 'd:%s' %(d2-d1)
                return True
        return False



#!/usr/bin/env python

# just a testbed to experiment with matplotlib, pyqt and so on, useful when changing version of one of them

import sys
import random

from PyQt5 import QtCore, QtWidgets

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.lines as mlines
import matplotlib.pyplot as plt


class MplCanvas(FigureCanvas):
    def __init__(self, parent=None, width=1, height=1, x_lower=0, x_upper=50, y_lower=1, y_upper=6, x_visible=True,
                 y_visible=True):
        figure = Figure(figsize=(width, height))
        self.x_lower, self.x_upper = [x_lower, x_upper]
        self.y_lower, self.y_upper = [y_lower, y_upper]
        self.x_visible, self.y_visible = [x_visible, y_visible]
        self.axes = figure.add_subplot(111)
        self.set_axis()

        FigureCanvas.__init__(self, figure)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)

        self.update_figure()
        # timer = QtCore.QTimer(self)
        # timer.timeout.connect(self.update_figure)
        # timer.start(1000)

    def set_axis(self):
        self.axes.axis('scaled')
        # We want the axes cleared every time plot() is called
        # self.axes.hold(False)
        self.axes.set_xbound(lower=self.x_lower-10, upper=self.x_upper)
        self.axes.set_ybound(lower=self.y_lower, upper=self.y_upper+5)
        self.axes.get_xaxis().set_visible(self.x_visible)
        self.axes.get_yaxis().set_visible(self.y_visible)

    def update_figure(self):
        self.axes.clear()
        self.set_axis()
        old_x, old_y = [self.x_lower, 0]
        for x in range(self.x_lower+1, self.x_upper):
            y = random.randint(self.y_lower, self.y_upper)
            line = mlines.Line2D([old_x, x], [old_y, y])
            self.axes.add_line(line)
            old_x, old_y = [x, y]

        show_arrow(self.axes, [0, 0], [5, 5])
        rectangle = plt.Rectangle((2, 2), 1, 10, angle=60, alpha=0.5)
        self.axes.add_patch(rectangle)

        self.draw()


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, win_title='Test window', width=1, height=1):
        QtWidgets.QMainWindow.__init__(self)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)  # Garbage-collect the window after it's been closed.
        self.setWindowTitle(win_title)

        main_widget = QtWidgets.QWidget(self)
        self.setCentralWidget(main_widget)

        box_layout = QtWidgets.QVBoxLayout(main_widget)

        canvas = MplCanvas(main_widget, width=width, height=height)
        box_layout.addWidget(canvas)

        self.show()


def show_arrow(ax, p1, p2, color='Black', connectionstyle='arc3,rad=0.3'):
    ax.annotate("", xy=p1, xycoords='data', xytext=p2, textcoords='data',
                arrowprops=dict(arrowstyle="->", color=color, shrinkA=5, shrinkB=5, patchA=None, patchB=None,
                                connectionstyle=connectionstyle,
                                ),
                )


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    title = 'Test if PtQt is working, should draw a graph and a rectangle'
    app_window = MainWindow(win_title=title, width=15, height=10)
    sys.exit(app.exec_())



from setuptools import setup, find_packages

setup(
    name='WinDeklar',
    version='0.1.7',
    description='Create winforms in an easy, declarative way. Specially suited for Robotics applications.',
    author='Nicolas Jodal',
    author_email='jnj@genexus.com',
    packages=find_packages(),
    install_requires=[
        # list your package dependencies here, for example:
        'numpy >= 1.15.0'
        'PyQt5~=5.14.1'
        'matplotlib~=3.5.2'
        'PyYAML~=5.3.1'
    ],
)



#!/usr/bin/env python

import sys
import time

import WinDeklar.WindowForm as WinForm
import WinDeklar.graph_aux as ga
import WinDeklar.QTAux as QTAux
import WinDeklar.record as rc
import WinDeklar.yaml_functions as ya


class ExampleHost(WinForm.HostModel):
    """
    Example of a Form with many features:
        * Open and Save files
        * Toolbar
        * Many kinds of controls (Slider, Combo, Check, Button, etc.)
        * Mouse move
    """

    def __init__(self, default_directory='/tmp', file_extension='yaml'):
        # keys (names used in the yaml definition file)
        self.check1_key = 'check1'
        self.points_key = 'points'
        self.axis_key   = 'show_axis'
        self.type_key   = 'graph_type'
        self.action_key = 'action'
        self.width_key  = 'line_width'

        # particular data
        self.action_name        = 'Action'
        self.last_action_number = 0
        self.directory          = default_directory
        self.file_extension     = file_extension
        self.file_filter        = '*.%s' % self.file_extension
        self.file_name          = None

        initial_values = {}  # used in case some control have an initial value programmatically
        super(ExampleHost, self).__init__(initial_values=initial_values)

    def update_view(self, ax):
        """
        Update the figure
        Notes:
            - This method is called when any property changes or refresh() is used
            - All form variables are in dict self.state
        :param ax:
        :return:
        """
        points = self.get_graph_points()
        show_axis = [True, True] if self.state.get(self.axis_key, True) else [False, False]
        ga.graph_points(ax, points, x_visible=show_axis[0], y_visible=show_axis[1],
                        line_width=self.state.get(self.width_key, 1.0))

    def redraw(self):
        """
        Event defined in the yaml file to be called when button redraw is pressed
        :return:
        """
        self.refresh()

    # actions
    def event_open_file(self):
        """
        Event defined in the yaml file to be called when File/Open is clicked
        :return:
        """
        file_name = self.get_file_name(title='Open an YAML', file_filter=self.file_filter, directory=self.directory)
        if file_name is None:
            return
        self.file_name = file_name
        # particular code to open the file
        self.open_yaml_file(self.file_name, progress_bar=self.get_progress_bar())

    def event_save_file_as(self):
        file_name = self.get_file_name_to_save(title='Save File', file_filter=self.file_filter,
                                               directory=self.directory)
        if file_name is None:
            return
        self.save_file(file_name, progress_bar=self.get_progress_bar())

    def event_save_file(self):
        if self.file_name is None:
            self.event_save_file_as()
        else:
            self.save_file(self.file_name, progress_bar=self.get_progress_bar())

    def change_action(self):
        self.last_action_number += 1
        value = '%s %s' % (self.action_name, self.last_action_number)
        self.set_and_refresh_control(self.action_key, value)

    def on_mouse_move(self, event, ax):
        if event.xdata is None or event.ydata is None:
            return
        self.show_status_bar_msg('x:%.2f y:%.2f' % (event.xdata, event.ydata))

    # particular code
    def open_yaml_file(self, file_name, progress_bar):
        """
        Example on how to use the progress bar
        :param file_name:
        :param progress_bar: use to give feedback about the opening process
        :return:
        """
        file = ya.get_yaml_file(file_name, must_exist=True, verbose=True)
        self.state.update(file['state'])
        print(self.state)
        self.refresh()

        # just an example of how to use the ProgressBar, no actually needed in this case
        progress_bar_example(progress_bar, max_value=100, inc=20, sleep_time=0.2)

        msg = '%s opened' % file_name
        self.show_status_bar_msg(msg)

    def save_file(self, file_name, progress_bar):
        """
        Example on how to use the progress bar
        :param file_name:
        :param progress_bar: use to give feedback about the opening process
        :return:
        """
        record = rc.Record(file_name, dir=None, add_time_stamp=False)
        record.write_ln('version: 1')  # just to avoid warnings with editing in pycharm
        record.write_group('state', self.state, level=0)

        # just an example of how to use the ProgressBar, no actually needed in this case
        progress_bar_example(progress_bar, max_value=100, inc=20, sleep_time=0.2)

        msg = '%s saved' % file_name
        self.show_status_bar_msg(msg)

    def get_graph_points(self):
        """
        Returns the points (x,y) to be graphed depending on the graph type and number of points
        :return:
        """
        function_name    = self.state.get(self.type_key, 'None')
        number_of_points = int(self.state.get(self.points_key, 10))
        points, msg      = ga.graph_points_for_many_functions(function_name, number_of_points)
        if points is None:
            self.show_status_bar_msg('%s not implemented' % function_name)
            points = []
        return points


def progress_bar_example(progress_bar, max_value=100, inc=20, sleep_time=0.2):
    progress_bar.set_max(max_value)
    for i in range(0, max_value, inc):
        progress_bar.set_value(i)
        time.sleep(sleep_time)


if __name__ == '__main__':
    app = QTAux.def_app()
    provider = ExampleHost()        # class to handle events
    WinForm.set_winform(__file__, provider)
    sys.exit(app.exec_())



import numpy as np
import random
import math
import matplotlib.lines as mlines

import WinDeklar.points_box as pb


class RealTimeDataProvider(object):
    def __init__(self, dt=0.1, min_y=0.0, max_y=10.0, color='Red'):
        self.min_y = min_y
        self.max_y = max_y
        self.color = color
        self.dt    = dt
        self.t     = 0.0

    def get_bounds(self):
        return self.min_y, self.max_y

    def get_min_y(self):
        return self.min_y

    def get_max_y(self):
        return self.max_y

    def get_next_values(self, i):
        x      = self.t
        self.t += self.dt
        y      = i
        return x, y


class RealTimeRandomDataProvider(RealTimeDataProvider):
    """
    Returns a random number between min_y and max_y
    """
    def __init__(self, dt=0.1, min_y=0.0, max_y=10.0, color='Red'):
        super(RealTimeRandomDataProvider, self).__init__(dt=dt, min_y=min_y, max_y=max_y, color=color)

    def get_next_values(self, i):
        x      = self.t
        self.t += self.dt
        return x, np.random.uniform(self.min_y, self.max_y)


class RealTimeFunctionDataProvider(RealTimeDataProvider):
    """
    Returns the result of applying a given function (function)
    """

    def __init__(self, dt=0.1, min_y=-1.2, max_y=1.2, function=np.sin, inc=np.radians(10), color='Red'):
        self.function = function
        self.inc      = inc
        self.last_r   = 0.0
        super(RealTimeFunctionDataProvider, self).__init__(dt=dt, min_y=min_y, max_y=max_y, color=color)

    def get_next_values(self, i):
        x      = self.t
        self.t += self.dt
        y = self.function(self.last_r)
        self.last_r += self.inc
        return x, y


class RealTimeConstantDataProvider(RealTimeDataProvider):
    def __init__(self, dt=0.1, min_y=0.0, max_y=10.0, color='Black'):
        self.reference = 0.0
        super(RealTimeConstantDataProvider, self).__init__(dt=dt, min_y=min_y, max_y=max_y, color=color)

    def set_reference(self, new_reference):
        self.reference = new_reference

    def get_next_values(self, i):
        x      = self.t
        self.t += self.dt
        return x, self.reference


def graph_points_for_many_functions(function_name, number_of_points):
    msg = ''
    if function_name == 'Random':
        points = random_function(0, number_of_points)
    elif function_name == 'Sine':
        points = get_function_xy_values(math.sin, 0, number_of_points, inc=math.radians(10))
    elif function_name == 'Cosine':
        points = get_function_xy_values(math.cos, 0, number_of_points, inc=math.radians(10))
    else:
        points = None
        msg    = '%s not implemented' % function_name
    return points, msg


def graph_points(ax, points, inc=1.1, scale_type='scaled', x_visible=True, y_visible=True, line_width=1.0,
                 color='Blue'):
    point_box = pb.PointsBox()
    point_box.add_points(points)  # to calc bounds
    ax.axis(scale_type)
    ax.get_xaxis().set_visible(x_visible)
    ax.get_yaxis().set_visible(y_visible)
    number_of_points = len(points)
    if number_of_points == 0:
        pass
    else:
        first_point = points[0]
        for i in range(1, number_of_points):
            [x2, y2] = points[i]
            line     = mlines.Line2D([first_point[0], x2], [first_point[1], y2], color=color, linewidth=line_width)
            ax.add_line(line)
            first_point = [x2, y2]
    point_box.set_bounds(ax, inc)


def random_function(from_x, to_x, min_y=0, max_y=10):
    # x = np.linspace(from_x, to_x)
    # y = np.random.randint(min_y, max_y, len(x))
    # return x, y
    return [[x, random.randint(min_y, max_y)] for x in range(from_x, to_x)]


def get_function_xy_values(function, from_x, to_x, inc=math.radians(5)):
    points = []
    x      = 0.0
    for i in range(from_x, to_x):
        points.append([x, function(x)])
        x += inc
    return points
